from collections import OrderedDict
from interpreter13 import NodeVisitor, BuiltinTypeSymbol, VarSymbol


class SemanticAnalyzer(NodeVisitor):
    def __init__(self):
        self.symbol_table = SymbolTable()

    def visit_Program(self, node):
        self.visit(node.block)

    def visit_Block(self, node):
        for declaration in node.declarations:
            self.visit(declaration)
        self.visit(node.compound_statement)

    def visit_VarDecl(self, node):
        symbol_type = node.type_node.name
        self.symbol_table.lookup(symbol_type)  # 从符号表查询内置类型符号

        var_name = node.var_node.name
        var_symbol = VarSymbol(var_name, symbol_type)
        self.symbol_table.insert(var_symbol)

    def visit_Compound(self, node):
        for child in node.children:
            self.visit(child)

    def visit_NoOperator(self, node):
        pass

    def visit_Variable(self, node):
        var_name = node.name
        var_symbol = self.symbol_table.lookup(var_name)

    def visit_BinaryOperator(self, node):
        self.visit(node.left)
        self.visit(node.right)

    def visit_Assign(self, node):
        self.visit(node.left)
        self.visit(node.right)


class SymbolTable:
    def __init__(self):
        self._symbols = OrderedDict()
        self._init_builtins()  # 初始化内置类型

    def _init_builtins(self):  # 定义初始化内置类型的方法
        self.insert(BuiltinTypeSymbol('INTEGER'))  # 通过insert()方法存入内置类型符号
        self.insert(BuiltinTypeSymbol('REAL'))  # 通过insert()方法存入内置类型符号

    def __str__(self):
        symtab_header = '符号表中的内容：'
        lines = [symtab_header, '-' * len(symtab_header) * 2]
        lines.extend([f'{key:8}: {value}' for key, value in self._symbols.items()])
        s = '\n'.join(lines)
        return s

    __repr__ = __str__

    def insert(self, symbol):
        print(f'存入：{symbol}')
        self._symbols[symbol.name] = symbol

    def lookup(self, name):  # 添加查询方法
        print(f'查询：{name}')
        symbol = self._symbols.get(name)
        return symbol


if __name__ == '__main__':
    from interpreter13 import Lexer, Parser

    text = '''
        program SymTab4;
            var x, y : integer;
        
        begin
            x := x + y;
        end.
    '''
    lexer = Lexer(text)
    parser = Parser(lexer)
    tree = parser.parser()
    semantic_analyzer = SemanticAnalyzer()
    semantic_analyzer.visit(tree)
    print(semantic_analyzer.symbol_table)
