from collections import OrderedDict
from interpreter13 import NodeVisitor, BuiltinTypeSymbol, VarSymbol


class SemanticAnalyzer(NodeVisitor):
    def __init__(self):
        self.symbol_table = SymbolTable()

    def visit_Program(self, node):
        self.visit(node.block)

    def visit_Block(self, node):
        for declaration in node.declarations:
            self.visit(declaration)
        self.visit(node.compound_statement)

    def visit_VarDecl(self, node):
        symbol_type = node.type_node.name
        self.symbol_table.lookup(symbol_type)

        var_name = node.var_node.name
        if self.symbol_table.lookup(var_name) is not None:  # 查询变量名称，如果存在变量信息
            raise Exception(f'错误：发现重复的标识符：{var_name}')  # 抛出异常
        var_symbol = VarSymbol(var_name, symbol_type)
        self.symbol_table.insert(var_symbol)

    def visit_Compound(self, node):
        for child in node.children:
            self.visit(child)

    def visit_NoOperator(self, node):
        pass

    def visit_Variable(self, node):
        var_name = node.name
        var_symbol = self.symbol_table.lookup(var_name)
        if var_symbol is None:
            raise NameError(f'错误：引用了不存在的标识符：{repr(var_name)}')

    def visit_BinaryOperator(self, node):
        self.visit(node.left)
        self.visit(node.right)

    def visit_Assign(self, node):
        self.visit(node.left)
        self.visit(node.right)


class SymbolTable:
    def __init__(self):
        self._symbols = OrderedDict()
        self._init_builtins()

    def _init_builtins(self):
        self.insert(BuiltinTypeSymbol('INTEGER'))
        self.insert(BuiltinTypeSymbol('REAL'))

    def __str__(self):
        symtab_header = '符号表中的内容：'
        lines = [symtab_header, '-' * len(symtab_header) * 2]
        lines.extend([f'{key:8}: {value}' for key, value in self._symbols.items()])
        s = '\n'.join(lines)
        return s

    __repr__ = __str__

    def insert(self, symbol):
        print(f'存入：{symbol}')
        self._symbols[symbol.name] = symbol

    def lookup(self, name):
        print(f'查询：{name}')
        symbol = self._symbols.get(name)
        return symbol


if __name__ == '__main__':
    from interpreter13 import Lexer, Parser

    text = '''
        program SymTab6;
           var x, y : integer;
           var y : real;
        begin
           x := x + y;
        end.
    '''
    lexer = Lexer(text)
    parser = Parser(lexer)
    tree = parser.parser()
    semantic_analyzer = SemanticAnalyzer()
    try:
        semantic_analyzer.visit(tree)
    except Exception as e:
        print(e)
    print(semantic_analyzer.symbol_table)
